#include "rumd_rdf.h"
#include "rumd/RUMD_Error.h"
#include <iostream>
#include <sstream>
#include <iomanip>

rumd_rdf::rumd_rdf() : directory("TrajectoryFiles"),
		       C0(),
		       numParticlesOfType(0),
		       rdfNonNorm(0),
		       sdfNonNorm(0),
		       rdf(0),
		       sdf(0),
		       rVals(0),
		       num_bins(0), 
		       L(0.),
		       dx(0.),
		       dt(0.),
		       num_part(0),
		       num_types(0),
		       MaxIndex(0),
		       Count(0),
		       verbose(false),
		       numFilenameDigits(4),
		       writeEachConfiguration(false) {}

void rumd_rdf::AllocateArrays(unsigned int nTypes, unsigned int nBins) {
  assert(nTypes > 0 && nBins > 0);

  if(nTypes != num_types || nBins != nBins) {
    FreeArrays();

    num_types = nTypes;
    num_bins = nBins;
   
    numParticlesOfType = new int[num_types];

    rdfNonNorm = new long**[num_types];
    rdf = new double**[num_types];
    for (unsigned int i=0; i<num_types; i++) {
      rdfNonNorm[i] = new long*[num_types];
      rdf[i] = new double*[num_types];
      for (unsigned int j=0; j<num_types; j++) {
        rdfNonNorm[i][j] = new long[num_bins];
        rdf[i][j] = new double[num_bins];
      }
    }
    
    sdfNonNorm = new long**[num_types];
    sdf = new double**[num_types];
    for (unsigned int i=0; i<num_types; i++) {
      sdfNonNorm[i] = new long*[DIM];
      sdf[i] = new double*[DIM];
      for (unsigned int j=0; j<DIM; j++) {
        sdfNonNorm[i][j] = new long[num_bins];
	sdf[i][j] = new double[num_bins];
      }
    }

    rVals = new double[num_bins];
  } // if(nTypes != ... )
}


void rumd_rdf::FreeArrays() {
  if(!num_types) return;

  delete [] numParticlesOfType; 

  for (unsigned int i=0; i<num_types; i++) {
    for (unsigned int j=0; j<num_types; j++) {
      delete [] rdfNonNorm[i][j];
      delete [] rdf[i][j];
    }
    delete [] rdfNonNorm[i];
    delete [] rdf[i];
  }
  delete [] rdfNonNorm;
  delete [] rdf;
  
  
  for (unsigned int i=0; i<num_types; i++) {
    for (unsigned int j=0; j<DIM; j++) {
      delete [] sdfNonNorm[i][j];
      delete [] sdf[i][j];
    }
    delete [] sdfNonNorm[i];
    delete [] sdf[i];
  }
  
  delete [] sdfNonNorm;
  delete [] sdf;

  delete [] rVals;
}


void rumd_rdf::ResetArrays() {
  for (unsigned int i=0; i<num_types; i++)
    for (unsigned int j=0; j<num_types; j++)
      for (int k=0; k<num_bins; k++)
	rdfNonNorm[i][j][k] = 0;
  
  
  for (unsigned int i=0; i<num_types; i++)
    for (unsigned int j=0; j<DIM; j++)
      for (int k=0; k<num_bins; k++)
	sdfNonNorm[i][j][k] = 0;
}

unsigned long int rumd_rdf::GetLastIndex() {
  gzFile gzfp;

  std::ostringstream LC0Name;
  LC0Name << directory << "/block0000.xyz.gz";

  gzfp=gzopen(LC0Name.str().c_str(),"r");
  unsigned long int lastIndex = 0;
  for(unsigned int j=0; j<=MaxIndex; j++)
    C0.Read(gzfp, LC0Name.str().c_str(), false);

  lastIndex = C0.metaData.timeStepIndex;
  gzclose(gzfp);

  return lastIndex;
}

void rumd_rdf::CalcSingleRDF() {
  unsigned int num_part = C0.num_part;

  double invDx = 1./dx;
  double    Lx = C0.metaData.L[0];
  double invLx = 1./Lx;
  double    Ly = C0.metaData.L[1];
  double invLy = 1./Ly;
  double    Lz = C0.metaData.L[2];
  double invLz = 1./Lz;
  int dr_index;

  for (unsigned int i=0; i<num_part; i++) {
    unsigned int TypeI =  C0.P[i].MyType;
    assert(TypeI<num_types);
    dr_index = (int)floor(( C0.P[i].x + Lx/2.)*invDx);
    if (dr_index < num_bins && dr_index >=0 ) sdfNonNorm[TypeI][0][dr_index]++;
    dr_index = (int)floor(( C0.P[i].y + Ly/2.)*invDx);
    if (dr_index < num_bins && dr_index >=0 ) sdfNonNorm[TypeI][1][dr_index]++;
    dr_index = (int)floor(( C0.P[i].z + Lz/2.)*invDx);
    if (dr_index < num_bins && dr_index >=0 ) sdfNonNorm[TypeI][2][dr_index]++;

    for (unsigned int j=0; j<i; j++) {
      int TypeJ =  C0.P[j].MyType;
      double R = sqrt( C0.P[i].RsqMinImage(&C0.P[j], Lx, Ly, Lz, invLx, invLy, invLz ) );
      dr_index = (int)floor(R*invDx);
      assert(dr_index >= 0);
      if (dr_index < num_bins) {
	rdfNonNorm[TypeI][TypeJ][dr_index]++;
	rdfNonNorm[TypeJ][TypeI][dr_index]++;
      }
    }
  }
}


void rumd_rdf::ComputeRDF(int nBins, float min_dt, unsigned int first_block, int last_block) {
  unsigned int last_saved_block;
  unsigned int last_block_to_read;
  unsigned int blockSize = 0;
  gzFile gzfp;
  FILE *fp;


  std::string lastComp_filename = directory+"/LastComplete_block.txt";

  fp=fopen(lastComp_filename.c_str(),"r");
  if (!fp)
    throw RUMD_Error("rumd_rdf","ComputeRDF",std::string("Error opening file ") + lastComp_filename + ": " + strerror(errno));
  fscanf(fp, "%u", &last_saved_block);
  int scan_blockSize = fscanf(fp, "%u", &blockSize);
  fclose(fp);

  if(scan_blockSize < 1)
    throw RUMD_Error("rumd_rdf", "ComputeRDF", "Old version of LastComplete_block.txt is missing the blockSize; please add it by hand");

  if(first_block > last_saved_block) {
    std::cerr << "Warning: user-specified first_block out of range; setting to zero" << std::endl;
    first_block = 0;
  }
  if(last_block < 0)
    last_block_to_read = last_saved_block;
  else
    if((unsigned int)last_block > last_saved_block) {
      last_block_to_read = last_saved_block;
      std::cerr << "Warning, specified last_block out of range; using last saved block" << std::endl;
    }
    else
      last_block_to_read = (unsigned int) last_block;


  if(first_block > last_block_to_read)
    std::cerr << "Warning: last_block smaller than first_block: no rdfs will be computed" << std::endl;

  // Read metadata from first configuration in first block
  std::ostringstream C0Name;
  C0Name << directory << "/block" << std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << first_block << ".xyz.gz";
  
  if(verbose)
    std::cout << "Reading block " << first_block << ": " << C0Name << std::endl << std::endl;

  gzfp=gzopen(C0Name.str().c_str(),"r");
  if (!gzfp)
    throw RUMD_Error("rumd_rdf","ComputeRDF",std::string("Error opening file") + C0Name.str() + ": " + strerror(errno) );

  C0.Read(gzfp, C0Name.str().c_str(), verbose);
  gzclose(gzfp);
 

  // Get meta-data and allocate arrays
  AllocateArrays(C0.metaData.numTypes, nBins); // sets num_types and num_bins

  L  = C0.metaData.L[0];     // Assuming Lx=Ly=Lz for now...
  assert(C0.metaData.L[0] == C0.metaData.L[1] && C0.metaData.L[0] == C0.metaData.L[2]); // make DIM-independent ...
  dx = L/1.0/float(num_bins);
  dt  = C0.metaData.dt;
  num_part = C0.num_part;
  MaxIndex = C0.metaData.logLin.maxIndex;
  

  ResetArrays();
  Count = 0;

  for (unsigned int tdx=0; tdx<num_types; tdx++)
    numParticlesOfType[tdx] = 0;
  for (unsigned int i=0; i<num_part; i++) {
    unsigned int Type = C0.P[i].MyType;
    assert(Type<num_types);
    numParticlesOfType[Type]++;
  }

  unsigned long int lastIndex = GetLastIndex();
  unsigned long int last_step_used;
  unsigned long int last_block_used;

  
  if(verbose) {
    for (unsigned int i=0; i<num_types; i++)
      std::cout << "Particles of type " << i << ": " << numParticlesOfType[i] << std::endl;
    std::cout << "LastIndex: " << lastIndex << std::endl;
    std::cout << "\nReading block: ";
  }
  
  for (unsigned int i=first_block; i<=last_block_to_read; i++) {

    // Read first configuration of block 'i'
    std::ostringstream C0Name;
    C0Name << directory << "/block" << std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << i << ".xyz.gz";

    if (verbose && (i-first_block)%10==0) std::cout << i << " ";
    fflush(stdout);
    gzfp=gzopen(C0Name.str().c_str(),"r");
    if (!gzfp)
      throw RUMD_Error("rumd_rdf","ComputeRDF",std::string("Error opening file ")+C0Name.str() + ": " + strerror(errno));

    C0.Read(gzfp,C0Name.str().c_str(),false);

    // Check that relevant metaData didn't change
    assert(C0.metaData.dt       == dt);
    assert(C0.num_part          == num_part);
    assert(C0.metaData.numTypes == num_types);
    assert(MaxIndex == C0.metaData.logLin.maxIndex);
    unsigned int nDigits = 5;

    Count++;
    CalcSingleRDF();
    if(writeEachConfiguration) {
      std::ostringstream rdf_filename;
      rdf_filename << "rdf_" << std::setfill('0') << std::setw(nDigits) << std::setprecision(nDigits) << i << "_" << std::setfill('0') << std::setw(nDigits) << std::setprecision(nDigits) << 0 << ".dat";
      Normalize();
      WriteRDF(rdf_filename.str());
      ResetArrays();
      Count = 0;
    }
    last_step_used = C0.metaData.timeStepIndex;
    last_block_used = i;

    // Analyze configurations in the same block-file
    for (unsigned int j=1; j<=MaxIndex; j++) {
      C0.Read(gzfp, C0Name.str().c_str(), false);
      
      // if ( (C0.metaData.timeStepIndex - last_step_used)*dt > min_dt && (lastIndex - C0.metaData.timeStepIndex)*dt > min_dt ){
      if ( ( (i - last_block_used) * blockSize 
	     +  C0.metaData.timeStepIndex - last_step_used)*dt > min_dt){
       	Count++;
	CalcSingleRDF();
	
	if(writeEachConfiguration) {
	  std::ostringstream rdf_filename;
	   rdf_filename << "rdf_" << std::setfill('0') << std::setw(nDigits) << std::setprecision(nDigits) << i << "_" << std::setfill('0') << std::setw(nDigits) << std::setprecision(nDigits) << j << ".dat";
	   Normalize();
	   WriteRDF(rdf_filename.str());
	   ResetArrays();
	   Count = 0;
	 }

	 last_step_used = C0.metaData.timeStepIndex;
	 last_block_used = i;
      }
    }
    gzclose(gzfp);
  } // end loop over blocks
  if(verbose)
    std::cout << std::endl;

  if(!writeEachConfiguration)
    Normalize();

}

void rumd_rdf::Normalize() {
  for (int k=0; k<num_bins; k++) {
    double r_start = k*dx;
    double r_last = (k+1)*dx;
    double rhoVol = 1./pow(L,3.) * 4./3.*M_PI*(pow(r_last,3.) - pow(r_start,3.));
    rVals[k] = (r_start + r_last)/2.;
    for (unsigned int i=0; i<num_types; i++ ) {
      for (unsigned int j=0; j<num_types; j++ ) {
	double factor = 1./float(Count)/rhoVol;
	if (i==j) factor /= float(numParticlesOfType[i]*(numParticlesOfType[i]-1));
	else factor /= float(numParticlesOfType[i]*numParticlesOfType[j]);
	rdf[i][j][k] = rdfNonNorm[i][j][k] * factor;
      } // j loop rdf
      
      for (unsigned int j=0; j<3; j++ ) {
	double factor = L/dx/float(Count)/float(num_part);
	sdf[i][j][k] = sdfNonNorm[i][j][k] * factor;
      } // j loop sdf
    } // i loop
  } // k loop
  
}


void rumd_rdf::WriteRDF(const std::string& filename) {

  FILE* fp = fopen(filename.c_str(),"w");

  fprintf(fp, "# concentrations:");
  for(unsigned int tdx=0; tdx < num_types; ++tdx)
    fprintf(fp, " %f", float(numParticlesOfType[tdx])/num_part );
  fprintf(fp, "\n");

  for (int k=0; k<num_bins; k++) {
    fprintf(fp, "%f ", rVals[k]);
     for (unsigned int i=0; i<num_types; i++ ) {
       for (unsigned int j=0; j<num_types; j++ )
	 fprintf(fp, "%f ", rdf[i][j][k]);
     }
     fprintf(fp, "\n");
  }
  fclose(fp);
}

void rumd_rdf::WriteSDF(const std::string& filename) {
  
  FILE* fp = fopen(filename.c_str(),"w");
  for (int k=0; k<num_bins; k++) {
     fprintf(fp, "%f ", rVals[k]);
     for (unsigned int i=0; i<num_types; i++ ) {
        for (int j=0; j<3; j++ ) {
	  fprintf(fp, "%f ", sdf[i][j][k]);
        }
     }
     fprintf(fp, "\n");
  }
  fclose(fp);
}
