
#ifndef MOLECULEDATA_H
#define MOLECULEDATA_H

#include "rumd/Sample.h"
#include "rumd/SimulationBox.h"
#include "rumd/PairPotential.h"
#include "rumd/ConstraintPotential.h"
#include "rumd/BondPotential.h"
#include "rumd/AnglePotential.h"
#include "rumd/DihedralPotential.h"

#include "cuda.h"
#include <float.h>

class MoleculeData {
  friend class BondHarmonic;
  friend class BondFENE;
  friend class AngleCosSq;
  friend class DihedralRyckaert;
  
 private:
  MoleculeData(const MoleculeData&); 
  MoleculeData& operator=(const MoleculeData&); 

  Sample* S;
  RectangularSimulationBox *SB;
  ConstraintPotential* constraintPotential;
 
  unsigned num_mol;         // Number of molecules 
  unsigned max_num_uau;     // Maximum nuau per molecule
  unsigned max_num_bonds;   // Maximum bonds per particle
  unsigned num_btypes;      // Number of bonding types 

  unsigned num_angles;      // Number of angles
  unsigned num_atypes;      // Number of angle types
  
  unsigned num_dihedrals;   // Number of dihedrals
  unsigned num_dtypes;      // Number of dihedral types

  unsigned flagHarmonic;    // Harmonic bond class instantiated
  unsigned flagFENE;        // FENE bond class instantiated
  unsigned flagConstraint;
  unsigned flagAngle;  
  unsigned flagDihedral;  

 public:
  MoleculeData(Sample *, const char *);
  ~MoleculeData(void);
  
  BondHarmonic *HarmBond;  
  BondFENE *FENEBond;
  AngleCosSq *AngleCos;
  DihedralRyckaert *DihedralRyc;


  // Bond Lists/arrays
  uint1 *h_blist,*d_blist;   // Bonding list 
  unsigned lblist; 

  int1 *h_btlist,*d_btlist;  // Bond type list 
  unsigned lbtlist;            
  
  int1 *h_btlist_int,*d_btlist_int;  // Internal bondlist

  float2 *h_bplist, *d_bplist; // Bond type parameter list 


  // Angle lists/arrays
  uint4 *h_alist, *d_alist; // Angle list: 
                            // x, y, z: Three indicies for the
                            // particles constituting the angle
                            // w: Angle type

  float2 *h_aplist, *d_aplist; // Angle type parameter list 
                               // x: force constant
                               // y: zero force angle

  float *h_epot_angle, *d_epot_angle; // Holds the potential energy for
                                      // each angle

  // Dihedral lists/arrays
  uint4 *h_dlist, *d_dlist;  // Dihedral list:
                             // x, y, z, w: Atom indicies for 

  uint1 *h_dtype, *d_dtype;  // Dihedral type list

  float *h_dplist, *d_dplist; // Parameter lists            

  float *h_epot_dihedral,     // Holds the potential energy for
    *d_epot_dihedral;         // each dihedral

  // Molecule lists (Convenient for data analysis)
  int1 *h_mlist;            
  int1 *d_mlist;            
  unsigned lmlist;          
  
  // Data arrays
  float *h_bonds, *d_bonds;              // The bonds 
  float *h_angles, *d_angles;            // The angles
  float *h_dihedrals, *d_dihedrals;      // The dihedrals

  float4 *h_cm, *d_cm;                   // Centre of mass + cm.w mass 
  float3 *h_vcm, *d_vcm;                 // Velocity centre of mass
  float3 *h_s, *d_s;                     // (Intrinsic) angular momentum
  float3 *h_omega, *d_omega;             // (Intrinsic) angular velocity
  float *h_inertia, *d_inertia;          // Moment of intertia (length=9*num_mols)
                                         // Is symmetric, but keep all element due to 
                                         // calculation of ang. velocity - row-wise matrix

  // GPU specifics
  dim3 threads_per_block;  
  dim3 num_blocks;

  // Helper functions 
  void AddmlistEntry(unsigned *, unsigned, unsigned, unsigned);
  void AddblistEntry(unsigned , unsigned); 
  void AddbtlistEntry(unsigned *, unsigned, unsigned, unsigned);
  void ReadTopology(const char *);
  void CheckRead(unsigned, unsigned, unsigned);

  // Set methods.
  void SetBondHarmonic(unsigned type, float lbond, float ks);
  void SetBondConstraint(unsigned type, float lbond);  
  void SetBondFENE(unsigned type);
  void SetAngleCosSq(unsigned type, float theta0, float ktheta);
  void SetDihedralRyckaert(unsigned type, float c0, float c1, float c2, float c3, 
			   float c4, float c5); // To incorp. in SWIG *sigh*
 
  void SetExclusionBond(PairPotential *potential, unsigned etype);
  void SetExclusionMolecule(PairPotential* potential);  
  void SetExclusionMolecule(PairPotential* potential, unsigned molindex);
  void SetExclusionAngle(PairPotential* potential);
  void SetExclusionDihedral(PairPotential* potential);
  void SetNumberOfConstraintIterations( unsigned numIte );

  void SetHarmonicBondConstant(unsigned type, float ks);
  void SetAngleConstant(unsigned type, float ktheta);
  void SetDihedralRyckaertConstants(unsigned type, float c0, float c1, float c2, 
				    float c3, float c4, float c5);

  // Get methods.
  unsigned GetNumberOfMolecules(){ return num_mol; };
  unsigned GetMaximumNumberOfBonds(){ return max_num_bonds; }
  unsigned GetNumberOfAngles(){ return num_angles; }
  unsigned GetNumberOfDihedrals(){ return num_dihedrals; }

  void GetBondsFromTopology(const char *str);
  void GetAnglesFromTopology(const char *str);
  void GetDihedralsFromTopology(const char *str);
  
  void GetDihedrals( void );
  void GetAngles( void );

  // Write or output methods
  void WriteMolConf(const char *fstr);
  void WriteMolxyz(const char *fstr, unsigned mol_id, const char mode[2]);
  void WriteConstraints( void );
  
  // Run time data evaluations
  void EvalMolMasses( void );
  void EvalCM( void ); 
  void EvalVelCM( void ); 
  void EvalSpin(void);

  // Debugging.
  void PrintLists(void);
  void PrintLists(unsigned i);
  void PrintMlist(void);
};

// Should of course be moved to a generic place
void GaussPivotBackCPU(float *x, float *b, float *A, unsigned nrc);
double Determinant3(float *A);

#endif // MOLECULEDATA_H
