#ifndef PAIRPOTENTIAL_H
#define PAIRPOTENTIAL_H

/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/RUMD_Error.h"
#include "rumd/Potential.h"
#include "rumd/NeighborList.h"
#include "rumd/SimulationBox.h"

#include <cstring>
#include <algorithm>
#include <iostream> 
#include <vector>
#include <sstream>

////////////////////////////////////////////////////////////////////////////////////////////////////////
// This header implements specialized Pair Potentials. Implement your potential here.
// DISCLAIMER: h_params[i][j] grows from 0 and upwards. Rcut_ij should always be the 0'th parameter, i.e., h_params[i][j][0].
////////////////////////////////////////////////////////////////////////////////////////////////////////

class PairPotential : public Potential{
public:
  enum CutoffMethod {NS, SP, SF};
private:
  PairPotential(const PairPotential&);
  PairPotential& operator=(const PairPotential&); 
  
  static std::vector<unsigned int> objectIndexList;

  static unsigned int GetNewObjectIndex();

  static void RemoveObjectIndex(unsigned int index);


  
  virtual void LaunchPE_Kernel() = 0;

 protected:
  NeighborList neighborList;
  RectangularSimulationBox* testRSB;
  LeesEdwardsSimulationBox* testLESB;
  
  unsigned int allocatedSize;
  CutoffMethod cutoffMethod;
  const unsigned int pp_objectIndex;

  float h_params[MaxNumTypes][MaxNumTypes][NumParam];
  float4* d_f_pe;
  float4* h_f_pe;

  void AllocatePE_Array(unsigned int nvp);
  
 public: 
  PairPotential( CutoffMethod cutoff_method ); 
  virtual ~PairPotential() { RemoveObjectIndex(pp_objectIndex); }

  void SetAllParams(float h_params[MaxNumTypes][MaxNumTypes][NumParam]);
  
  // Needed due to no virtual function on device.
  virtual float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return 0.f; };  
  void WritePotentials(SimulationBox* simBox);
  void CopyParamsToGPU( unsigned i, unsigned j, unsigned usedParams ); 
  void SaveNeighborList(){ neighborList.SaveNeighborList(); };
  void RestoreNeighborList(){ neighborList.RestoreNeighborList(); };
  void ResetNeighborList(){ neighborList.ResetNeighborList(); };
  void PrintExclusionList(){ neighborList.PrintExclusionList(); };

  // Set methods.
  void SetNbListSkin( float skin ){ neighborList.SetCutoffs(this, skin); }
  void SetSampleParameters( KernelPlan& kPlan, ParticleData* pd, SimulationBox* simBox, float4* d_misc );
  void SetExclusionList( NBTYPEv* list ){ neighborList.SetExclusionList(list); };
  void SetExclusionBond( int1 *h_btlist, uint1* h_blist, unsigned num_part, unsigned num_virt_part, unsigned max_num_bonds, unsigned etype );
  void SetExclusionType( unsigned type0, unsigned type1 );
  void SetExclusionMolecule( int1 *h_mlist, unsigned molindex, unsigned max_num_uau );
  void SetExclusionDihedral(uint4 *h_dlist,  unsigned num_dihedrals);
  void SetExclusionAngle(uint4 *h_alist,  unsigned num_angles);

   // Get methods.
  float GetPotentialParameter(int type1, int type2, int paramIdx) const{ return h_params[type1][type2][paramIdx]; }
  float GetNbListSkin(){ return neighborList.GetSkin(); }
  double GetPotentialEnergy();
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Standard 12-6 Lennard-Jones
// v(r) = 4 Epsilon ( (Sigma/r)^12 - (Sigma/r)^6 )
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_LJ_12_6 : public PairPotential{
 public: 
  Pot_LJ_12_6( CutoffMethod cutoff_method ) : PairPotential(cutoff_method) { SetID_String("potLJ_12_6");}
  
  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();  
  
  Potential* Copy() {
    Pot_LJ_12_6* new_pot = new Pot_LJ_12_6(cutoffMethod);
    new_pot->SetAllParams(h_params);
    return new_pot;
  }
  
  void SetParams( unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut ){
    h_params[i][j][0] = Rcut * Sigma;            // Rcut_ij
    h_params[i][j][1] = 1.f / ( Sigma * Sigma ); // (1/Sigma_ij)^2
    h_params[i][j][2] = 48.f * Epsilon;          // Prefactor for forces
    h_params[i][j][3] = 4.f * Epsilon;         

    if(verbose)
      std::cout << GetID_String() + ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,4);
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f){
    float invDist2 = 1.0f / ( dist2 * param[1] );                              // (sigma/r)^2. 
    float invDist6 = invDist2 * invDist2 * invDist2;                           // (sigma/r)^6. 
    float s = param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 0.5f ); // F_ij = s * r_ij.
    (*my_f).w += param[3] * invDist6 * ( invDist6 - 1.0f );                    
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Generalized Lennard-Jones
// v(r) = Epsilon / (m -n) ( n*(Sigma/r)^m -  m*(Sigma/r)^n )
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_gLJ_m_n : public PairPotential{
 public: 
  float m, n; // exponents 
  
  Pot_gLJ_m_n(float set_m, float set_n, CutoffMethod cutoff_method ): PairPotential(cutoff_method), m(set_m), n(set_n){ std::stringstream ss; ss << "potgLJ_" << m << "_" << n; SetID_String(ss.str()); }
  
  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();
  
  Potential* Copy() {
    Pot_gLJ_m_n* new_pot = new Pot_gLJ_m_n(m, n, cutoffMethod);
    new_pot->SetAllParams(h_params);
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    h_params[i][j][0] = Rcut * Sigma;            // Rcut_ij
    h_params[i][j][1] = 1.f / ( Sigma * Sigma ); // (1/Sigma_ij)^2
    h_params[i][j][2] = m*n / (m-n) * Epsilon;   // Prefactor for forces
    h_params[i][j][3] = m*0.5f;                  // m/2.0
    h_params[i][j][4] = n*0.5f;                  // n/2.0
    h_params[i][j][5] = Epsilon / (m-n);   
     
    if(verbose)
      std::cout << GetID_String() << ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,6);
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f) {
    float invDist2 = 1.0f / ( dist2 * param[1] );                             
    float invDist_m = exp(log(invDist2)*param[3]); 
    float invDist_n = exp(log(invDist2)*param[4]);
    float s =   param[1] * param[2] * (invDist_m - invDist_n) * invDist2; 
    (*my_f).w += param[5] * ( 2.0f*param[4]*invDist_m - 2.0f*param[3]*invDist_n );
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Standard 12-6 Lennard-Jones (re-def. Sigma HO = 2^(1/6) Sigma LJ)
// v(r) = Epsilon ( (Sigma/r)^12 -  2 *  (Sigma/r)^6 )
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_LJ_12_6_NEW : public PairPotential{
 public: 
  Pot_LJ_12_6_NEW( CutoffMethod cutoff_method ) : PairPotential(cutoff_method) { SetID_String("potLJN"); }

  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();

  Potential* Copy() {
    Pot_LJ_12_6_NEW * new_pot = new Pot_LJ_12_6_NEW (cutoffMethod);
    new_pot->SetAllParams(h_params);
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    h_params[i][j][0] = Rcut * Sigma;             // Rcut_ij
    h_params[i][j][1] = 1.f / ( Sigma * Sigma ); // (1/Sigma_ij)^2
    h_params[i][j][2] = 12.f * Epsilon;           // Prefactor for forces
    h_params[i][j][3] = Epsilon;         

    
    if(verbose)
      std::cout << GetID_String() << ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,3);
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f){
    float invDist2 = 1.0f / ( dist2 * param[1] );                                
    float invDist6 = invDist2 * invDist2 * invDist2;                             
    float s =   param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 1.f );  
    (*my_f).w += param[3] * invDist6 * ( invDist6 - 2.0f );          
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// 12-6 Lennard-Jones plus Gaussian potential
// v(r) = Epsilon ( (Sigma/r)^12 -  2(Sigma/r)^6 - Epsilon_0 (exp[-(r - r_0)^2 / (2 Sigma_0^2)]) ) 
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_LJ_12_6_Gauss : public PairPotential{
 public: 
  Pot_LJ_12_6_Gauss( CutoffMethod cutoff_method ) : PairPotential(cutoff_method) { SetID_String("potLJG"); }

  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();
  
  Potential* Copy() {
    Pot_LJ_12_6_Gauss* new_pot = new Pot_LJ_12_6_Gauss(cutoffMethod);
    new_pot->SetAllParams(h_params);
    return new_pot;  
  }
  
  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut, float Sigma0, float Epsilon0, float r0) {
    h_params[i][j][0] = Rcut * Sigma;            
    h_params[i][j][1] = 1.f / ( Sigma * Sigma ); 
    h_params[i][j][2] = 12.f * Epsilon;          
    h_params[i][j][3] = 2.0f * Sigma0 * Sigma0;  
    h_params[i][j][4] = r0;                      
    h_params[i][j][5] = Epsilon0;                
    h_params[i][j][6] = Epsilon;                

    if(verbose)
      std::cout << GetID_String() << ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Sigma0 = " << Sigma0 
		<< ", Epsilon0 = " << Epsilon0 << ", r0 = " << r0 << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,7);
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f) {
    float invDist2 = 1.0f / ( dist2 * param[1] );                             
    float invDist6 = invDist2 * invDist2 * invDist2;                          
    float r = sqrtf(dist2);
    float exp_factor = exp (- ((r - param[4]) * (r - param[4])) / param[3]);
    float s = param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 1.f ) - (param[5] * param[2] * ( r - param[4] ) / ( 6.0f * r *  param[3])) * exp_factor; 
    (*my_f).w += param[6] * (invDist6 * ( invDist6 - 2.0f ) - param[5] * exp_factor );                                                 
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Gaussian core potential.
// v(r) = Epsilon * exp[-(r/Sigma)^2]
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_Gauss : public PairPotential{
 public: 
  Pot_Gauss( CutoffMethod cutoff_method ) : PairPotential(cutoff_method) { SetID_String("potGau"); }

  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();

  Potential* Copy() {
    Pot_Gauss* new_pot = new Pot_Gauss(cutoffMethod);
    new_pot->SetAllParams(h_params);
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    h_params[i][j][0] = Rcut * Sigma;           
    h_params[i][j][1] = 1.f / ( Sigma * Sigma ); 
    h_params[i][j][2] = Epsilon;                 

    if(verbose)
      std::cout << GetID_String() << ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,3);
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f){
    float Dist2 = dist2 * param[1];                    // (r/Sigma)^2 
    float s = 2.f * param[1] * param[2] * exp(-Dist2); // F_ij = s * r_ij
    (*my_f).w += param[2] * exp(- Dist2);              // U_ij 
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Buckingham potential.
// v(r) = Epsilon{(6/(Alpha - 6))*exp[Alpha(1-r/rm)] - (Alpha/(Alpha -6))*(rm/r)**6}
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_Buckingham : public PairPotential{
 public: 
  Pot_Buckingham( CutoffMethod cutoff_method ) : PairPotential(cutoff_method) { SetID_String("potBuc"); }
  
  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();

  Potential* Copy() {
    Pot_Buckingham* new_pot = new Pot_Buckingham(cutoffMethod);
    new_pot->SetAllParams(h_params);
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float rm, float Alpha, float Epsilon, float Rcut){
    h_params[i][j][0] = Rcut;  
    h_params[i][j][1] = 1.f / (rm * rm);                    // (1/rm_ij)^2
    h_params[i][j][2] = Alpha;
    h_params[i][j][3] = 6.f * Epsilon / (Alpha - 6.f);      // 6 / (Alpha -6)
    h_params[i][j][4] = Alpha * Epsilon / (Alpha - 6.f);    // Alpha/(Alpha -6)
    h_params[i][j][5] = 6.f * Epsilon * Alpha / (Alpha -6); // Prefactor for s
    
    if(verbose)
      std::cout << GetID_String() << ": " << "Alpha = " << Alpha << ", Epsilon = " << Epsilon << ", rm = " << rm << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,5);
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f){
    float invDist2 = 1.0f / ( dist2 * param[1] );           // (rm/r)^2
    float invDist6 = invDist2 * invDist2 * invDist2;        // (rm/r)^6
    float invDist  = sqrt(invDist2);                        //  rm/r
    float expDist  = exp(param[2] * (1.f - 1.f/invDist));   // exponential term 

    float s = param[1]*param[5] * ( invDist*expDist - invDist6*invDist2 );
    (*my_f).w +=  param[3] * expDist - param[4] * invDist6;
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// IPL potential, n=12
// v(r) = Epsilon ( (Sigma/r)^12 )
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_IPL_12 : public PairPotential{
 public: 
  Pot_IPL_12( CutoffMethod cutoff_method ) : PairPotential(cutoff_method) { SetID_String("potIPL12"); }
  
  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();
  
  Potential* Copy() {
    Pot_IPL_12* new_pot = new Pot_IPL_12(cutoffMethod);
       new_pot->SetAllParams(h_params);
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    h_params[i][j][0] = Rcut * Sigma;             // Rcut_ij
    h_params[i][j][1] =  1.f / ( Sigma * Sigma ); // (1/Sigma_ij)^2
    h_params[i][j][2] = 12.f * Epsilon;           // Prefactor for forces
    h_params[i][j][3] = Epsilon;             

    if(verbose)
      std::cout << GetID_String() << ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,4);
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f) {
    float invDist2  = 1.0f / ( dist2 * param[1] );        // (sigma/r)^2 
    float invDist6  = invDist2 * invDist2 * invDist2;     // (sigma/r)^6
    float invDist12 = invDist6 * invDist6;                // (sigma/r)^12    
    float s = param[1] * param[2] * invDist12 * invDist2; // F_ij = s * r_ij
    (*my_f).w += param[3] * invDist12;                  
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// IPL potential, n=18
// v(r) = Epsilon ( (Sigma/r)^18 ) 
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_IPL_18 : public PairPotential{
 public: 
  Pot_IPL_18( CutoffMethod cutoff_method ) : PairPotential(cutoff_method) { SetID_String("potIPL18"); }

  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();  

  Potential* Copy() {
    Pot_IPL_18* new_pot = new Pot_IPL_18(cutoffMethod);
       new_pot->SetAllParams(h_params);
       return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    h_params[i][j][0] = Rcut * Sigma;            // Rcut_ij
    h_params[i][j][1] = 1.f / ( Sigma * Sigma ); // (1/Sigma_ij)^2
    h_params[i][j][2] = 18.f * Epsilon;          // Prefactor for forces
    h_params[i][j][3] = Epsilon;          
    
    if(verbose)
      std::cout << GetID_String() << ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,4);
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f) {
    float invDist2  = 1.0f / ( dist2 * param[1] );        // (sigma/r)^2 
    float invDist6  = invDist2 * invDist2 * invDist2;     // (sigma/r)^6
    float invDist18 = invDist6 * invDist6 * invDist6;     // (sigma/r)^18    
    float s = param[1] * param[2] * invDist18 * invDist2; // F_ij = s * r_ij
    (*my_f).w += param[3] * invDist18;                   
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// IPL potential, abritrary exponent n
// v(r) =   Epsilon ( (Sigma/r)^n )
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_IPL_n : public PairPotential{
 public: 
  float n;   // exponent
  
  Pot_IPL_n( float set_n, CutoffMethod cutoff_method ) : PairPotential(cutoff_method), n(set_n) { std::stringstream ss; ss << "potIPL_" << n; SetID_String(ss.str()); }

  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();

  Potential* Copy() {
    Pot_IPL_n* new_pot = new Pot_IPL_n(n, cutoffMethod);
    new_pot->SetAllParams(h_params);
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    h_params[i][j][0] = Rcut * Sigma;             // Rcut_ij
    h_params[i][j][1] =  1.f / ( Sigma * Sigma ); // (1/Sigma_ij)^2
    h_params[i][j][2] = Epsilon;                  // Prefactor for forces
    h_params[i][j][3] = n;                        // n

    if(verbose)
      std::cout << GetID_String() << ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,4);
  }

  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f) {
    float invDist2   = 1.0f / ( dist2 * param[1] );                  // (sigma/r)^2 
    float invDist_n = expf(logf(invDist2)*param[3]*0.5f);            // (sigma/r)^n (faster)
    float s = param[1] * param[2] * param[3] * invDist_n * invDist2; // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist_n;                              
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Integral of IPLs r^(-n) for n from 6 to inf
// v(r) = Epsilon ( (Sigma/r)^6 / ln(r/Sigma) )
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_IPL6_LN : public PairPotential{
 public: 
  Pot_IPL6_LN( CutoffMethod cutoff_method ) : PairPotential(cutoff_method) { SetID_String("potIP6LN");}
  
  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();  
  
  Potential* Copy() {
    Pot_IPL6_LN* new_pot = new Pot_IPL6_LN(cutoffMethod);
    new_pot->SetAllParams(h_params);
    return new_pot;
  }
  
  void SetParams( unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut ){
    h_params[i][j][0] = Rcut * Sigma;                 // Rcut_ij
    h_params[i][j][1] = Epsilon;                      // Prefactor 
    h_params[i][j][2] = 1.f / ( Sigma * Sigma );      // (1/Sigma_ij)^2
    h_params[i][j][3] = Epsilon / ( Sigma * Sigma ) ; // Prefactor for force
    
    if(verbose)
      std::cout << GetID_String() + ": " << "Sigma = " << Sigma << ", Epsilon = " << Epsilon << ", Rcut = " << Rcut << "." << std::endl;
    CopyParamsToGPU(i,j,4);
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f){
    float  lnDist  = 0.5f * log( dist2 * param[2] );  // ln(sigma/r).
    float invDist2 =    1.0f / ( dist2 * param[2] );  // (sigma/r)^2.
    float invDist6 = invDist2 * invDist2 * invDist2;  // (sigma/r)^6.
    
    float s = param[3] * invDist6 * invDist2 * ( 6.0f * lnDist + 1 ) / ( lnDist * lnDist );
    (*my_f).w += param[1] * invDist6 / lnDist;
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Dzugutov potential.
// v(r) = V1 + V2 = [ A ( r^(-n) - B ) * exp( c / ( r - a ) ) ] + [ B * exp( d / ( r - b ) ], V1 = 0 for r >= a 
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_Dzugutov : public PairPotential{
 public: 
  Pot_Dzugutov() : PairPotential(NS) { SetID_String("potDzu"); }

  void CalcF(bool initialize, bool calc_stresses);
  float ComputeInteraction_CPU(float dist2, float* param, float4* my_f) { return ComputeInteraction(dist2, param, my_f); };  
  void LaunchPE_Kernel();

  Potential* Copy() {
    Pot_Dzugutov* new_pot = new Pot_Dzugutov();
    new_pot->SetAllParams(h_params);
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float a, float b, float c, float d, float A, float B){
    h_params[i][j][0] = (b - 0.000001f); // To avoid division by zero.
    h_params[i][j][1] = a; 
    h_params[i][j][2] = b; 
    h_params[i][j][3] = c; 
    h_params[i][j][4] = d; 
    h_params[i][j][5] = A; 
    h_params[i][j][6] = B; 
    
    if(verbose)
      std::cout << GetID_String() << ": " << "a = " << a << ", b = " << b << ", c = " << c << ", d = " << d << ", A = " << A 
		<< ", B = " << B << "." << std::endl;
    CopyParamsToGPU(i,j,7);
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f){
    float r = sqrtf(dist2); 
    
    float inv_r = 1.f / r;
    float inv_b = 1.f / ( r - param[2] );
    float exp_b = expf( param[4] * inv_b );
    float inv_b_sq_exp = inv_b * inv_b * exp_b;
    
    float s = inv_r * param[4] * param[6] * inv_b_sq_exp; 
    (*my_f).w += param[6] * exp_b; 
    
    if( r < param[1] ){
      float invDist2 = 1.f / dist2; 
      float invDist4 = invDist2 * invDist2;
      float invDist8 = invDist4 * invDist4;
      
      float inv_a = 1.f / ( r - param[1] );
      float exp_a = expf( param[3] * inv_a );
      
      float potential_energy = param[5] * exp_a * ( invDist8 * invDist8 - param[6] );
      s += 16 * param[5] * (exp_a * invDist8 * invDist8) * invDist2 + inv_r * param[3] * inv_a * inv_a * potential_energy;
      (*my_f).w += potential_energy;
    }
    return s;
  }
};

#endif // PAIRPOTENTIAL_H
